from math import sqrt

import pytest

import moyopy


@pytest.fixture
def wurtzite() -> moyopy.Cell:
    # https://next-gen.materialsproject.org/materials/mp-560588
    a = 3.81
    c = 6.24
    basis = [
        [a, 0.0, 0.0],
        [-a / 2.0, a * sqrt(3.0) / 2.0, 0.0],
        [0.0, 0.0, c],
    ]
    z1_2b = 0.00014
    z2_2b = 0.37486
    positions = [
        # 2b
        [1 / 3, 2 / 3, z1_2b],
        [2 / 3, 1 / 3, z1_2b + 0.5],
        # 2b
        [1 / 3, 2 / 3, z2_2b],
        [2 / 3, 1 / 3, z2_2b + 0.5],
    ]
    numbers = [1, 1, 2, 2]

    cell = moyopy.Cell(basis, positions, numbers)
    return cell



class Cell:
    def __init__(
        self,
        basis: list[list[float]],
        positions: list[list[float]],
        numbers: list[int],
    ): ...
    @property
    def basis(self) -> list[list[float]]: ...
    @property
    def positions(self) -> list[list[float]]: ...
    @property
    def numbers(self) -> list[int]: ...
    def serialize_json(self) -> str: ...
    @classmethod
    def deserialize_json(cls, json_str: str) -> Cell: ...

class Operations:
    @property
    def rotations(self) -> list[list[list[float]]]: ...
    @property
    def translations(self) -> list[list[float]]: ...
    @property
    def num_operations(self) -> int: ...
    def __len__(self) -> int: ...



from __future__ import annotations

import numpy as np

from moyopy import Operations, operations_from_number


def _unique_sites_in_cell(position, operations: Operations) -> np.ndarray:
    def _is_close(site1, site2):
        diff = site1 - site2
        diff -= np.round(diff)
        return np.allclose(diff, 0, atol=1e-4)

    sites = []
    for rotation, translation in zip(operations.rotations, operations.translations):
        new_position = np.array(rotation) @ np.array(position) + np.array(translation)
        if np.any([_is_close(site, new_position) for site in sites]):
            continue
        sites.append(new_position)
    return np.array(sites)


def test_operations_from_number():
    operations = operations_from_number(number=230)  # Ia-3d
    num_operations = 48 * 2
    assert operations.num_operations == num_operations
    assert len(operations) == num_operations

    assert len(_unique_sites_in_cell([1 / 8, 1 / 8, 1 / 8], operations)) == 16



# -- Project information -----------------------------------------------------
from moyopy import __version__ as version

project = "moyopy"
copyright = "2023, Kohei Shinohara"
repository_url = "https://github.com/spglib/moyo"

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.bibtex",
    # "nbsphinx",
    "myst_parser",
    "autoapi.extension",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]

# The suffix(es) of source filenames.
source_suffix = {
    ".md": "markdown",
    ".rst": "restructuredtext",
}

# -----------------------------------------------------------------------------
# napoleon
# -----------------------------------------------------------------------------

# napoleon_type_aliases = {}
napoleon_use_rtype = True
napoleon_use_ivar = True

# -----------------------------------------------------------------------------
# bibtex
# -----------------------------------------------------------------------------

# https://pypi.org/project/sphinxcontrib-bibtex/
bibtex_bibfiles = ["references.bib"]

# -----------------------------------------------------------------------------
# intersphinx
# -----------------------------------------------------------------------------
intersphinx_mapping = {
    "spglib": ("https://spglib.readthedocs.io/en/stable/", None),
}

# -----------------------------------------------------------------------------
# MyST
# -----------------------------------------------------------------------------
myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "tasklist",
]
myst_dmath_double_inline = True
myst_heading_anchors = 3

# -----------------------------------------------------------------------------
# autoapi
# -----------------------------------------------------------------------------
autoapi_dirs = [
    "../python/moyopy",
]

# -----------------------------------------------------------------------------
# sphinx-book-theme
# -----------------------------------------------------------------------------
html_theme = "sphinx_book_theme"
html_title = project + " " + version
html_theme_options = {
    "repository_url": repository_url,
    "use_repository_button": True,
    "navigation_with_keys": True,
    "globaltoc_includehidden": "true",
    "show_toc_level": 3,
}

# hide sphinx footer
html_show_sphinx = False
html_show_sourcelink = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



from __future__ import annotations

import numpy as np
import pytest

import moyopy
from moyopy.interface import MoyoAdapter


def test_pymatgen_moyopy(wurtzite: moyopy.Cell):
    structure = MoyoAdapter.get_structure(wurtzite)
    cell = MoyoAdapter.from_structure(structure)
    assert np.allclose(cell.basis, wurtzite.basis)
    assert np.allclose(cell.positions, wurtzite.positions)
    assert cell.numbers == wurtzite.numbers


def test_ase_moyopy(wurtzite: moyopy.Cell):
    atoms = MoyoAdapter.get_atoms(wurtzite)
    cell = MoyoAdapter.from_atoms(atoms)
    assert np.allclose(cell.basis, wurtzite.basis)
    assert np.allclose(cell.positions, wurtzite.positions)
    assert cell.numbers == wurtzite.numbers


def test_from_py_obj(wurtzite: moyopy.Cell):
    # Create different structure types
    structure = MoyoAdapter.get_structure(wurtzite)
    atoms = MoyoAdapter.get_atoms(wurtzite)
    mson_atoms = structure.to_ase_atoms()

    # Test conversion from each type
    for struct in [structure, atoms, mson_atoms]:
        cell = MoyoAdapter.from_py_obj(struct)
        assert np.allclose(cell.basis, wurtzite.basis)
        assert np.allclose(cell.positions, wurtzite.positions)
        assert cell.numbers == wurtzite.numbers

    # Test invalid input type
    with pytest.raises(TypeError, match="Expected Structure, Atoms, or MSONAtoms, got list"):
        MoyoAdapter.from_py_obj([1, 2, 3])



from __future__ import annotations

from moyopy import HallSymbolEntry


def test_hall_symbol_entry():
    hs = HallSymbolEntry(hall_number=528)  # No. 228, origin choice 2
    assert hs.hall_number == 528
    assert hs.number == 228
    assert hs.setting == "2"
    assert hs.hm_short == "F d -3 c"



from moyopy._base import Cell, Operations  # noqa: F401
from moyopy._data import (
    Centering,
    HallSymbolEntry,
    Setting,
    SpaceGroupType,
    operations_from_number,
)  # noqa: F401

__version__: str

class MoyoDataset:
    """A dataset containing symmetry information of the input crystal structure."""
    def __init__(
        self,
        cell: Cell,
        symprec: float = 1e-4,
        angle_tolerance: float | None = None,
        setting: Setting | None = None,
    ):
        """
        Parameters
        ----------
        cell: Cell
            Input crystal structure.
        symprec: float
            Symmetry search tolerance in the unit of cell.lattice.
        angle_tolerance: float | None
            Symmetry search tolerance in the unit of radians.
        setting: Setting | None
            Preference for the setting of the space group.
        """
    # Space-group type
    @property
    def number(self) -> int:
        """Space group number."""
    @property
    def hall_number(self) -> int:
        """Hall symbol number."""
    # Symmetry operations in the input cell
    @property
    def operations(self) -> Operations:
        """Symmetry operations in the input cell."""
    # Site symmetry
    @property
    def orbits(self) -> list[int]:
        """Spglib's `crystallographic_orbits` not `equivalent_atoms`.

        The `i`th atom in the input cell is equivalent to the `orbits[i]`th atom in the **input**
        cell. For example, orbits=[0, 0, 2, 2, 2, 2] means the first two atoms are equivalent
        and the last four atoms are equivalent to each other.
        """
    @property
    def wyckoffs(self) -> list[str]:
        """Wyckoff letters for each site in the input cell."""
    @property
    def site_symmetry_symbols(self) -> list[str]:
        """Site symmetry symbols for each site in the input cell.

        The orientation of the site symmetry is w.r.t. the standardized cell.
        """
    # Standardized cell
    @property
    def std_cell(self) -> Cell:
        """Standardized cell."""
    @property
    def std_linear(self) -> list[list[float]]:
        """Linear part of transformation from the input cell to the standardized cell."""
    @property
    def std_origin_shift(self) -> list[float]:
        """Origin shift of transformation from the input cell to the standardized cell."""
    @property
    def std_rotation_matrix(self) -> list[list[float]]:
        """Rigid rotation."""
    @property
    def pearson_symbol(self) -> str:
        """Pearson symbol for standardized cell."""
    # Primitive standardized cell
    @property
    def prim_std_cell(self) -> Cell:
        """Primitive standardized cell."""
    @property
    def prim_std_linear(self) -> list[list[float]]:
        """Linear part of transformation from the input cell to the primitive standardized cell."""
    @property
    def prim_std_origin_shift(self) -> list[float]:
        """Origin shift of transformation from the input cell to the primitive standardized
        cell."""
    @property
    def mapping_std_prim(self) -> list[int]:
        """Mapping sites in the input cell to those in the primitive standardized cell.

        The `i`th atom in the input cell is mapped to the `mapping_to_std_prim[i]`th atom in the
        primitive standardized cell.
        """
    # Final parameters
    @property
    def symprec(self) -> float:
        """Actually used `symprec` in iterative symmetry search."""
    @property
    def angle_tolerance(self) -> float | None:
        """Actually used `angle_tolerance` in iterative symmetry search."""

__all__ = [
    # base
    "Cell",
    "Operations",
    # data
    "Setting",
    "Centering",
    "HallSymbolEntry",
    "SpaceGroupType",
    "operations_from_number",
    # lib
    "__version__",
    "MoyoDataset",
]



from __future__ import annotations

try:
    import ase
    from pymatgen.core import Element, Structure
    from pymatgen.io.ase import MSONAtoms
except ImportError:
    raise ImportError("Try installing dependencies with `pip install moyopy[interface]`")

import moyopy


class MoyoAdapter:
    @staticmethod
    def get_structure(cell: moyopy.Cell) -> Structure:
        species = [Element.from_Z(number) for number in cell.numbers]
        return Structure(lattice=cell.basis, species=species, coords=cell.positions)

    @staticmethod
    def get_atoms(cell: moyopy.Cell) -> ase.Atoms:
        atoms = ase.Atoms(
            cell=cell.basis,
            scaled_positions=cell.positions,
            numbers=cell.numbers,
            pbc=True,
        )
        return atoms

    @staticmethod
    def from_structure(structure: Structure) -> moyopy.Cell:
        basis = structure.lattice.matrix
        positions = structure.frac_coords
        numbers = [site.specie.Z for site in structure]

        return moyopy.Cell(
            basis=basis.tolist(),
            positions=positions.tolist(),
            numbers=numbers,
        )

    @staticmethod
    def from_atoms(atoms: ase.Atoms) -> moyopy.Cell:
        basis = list(atoms.cell)
        positions = atoms.get_scaled_positions()
        numbers = atoms.get_atomic_numbers()

        return moyopy.Cell(
            basis=basis,
            positions=positions,
            numbers=numbers,
        )

    @staticmethod
    def from_py_obj(struct: Structure | ase.Atoms | MSONAtoms) -> moyopy.Cell:
        """Convert a Python atomic structure object to a Moyo Cell.

        Args:
            struct: Currently supports pymatgen Structure, ASE Atoms, and MSONAtoms

        Returns:
            moyopy.Cell: The converted Moyo cell
        """
        if isinstance(struct, (ase.Atoms, MSONAtoms)):
            return MoyoAdapter.from_atoms(struct)
        elif isinstance(struct, Structure):
            return MoyoAdapter.from_structure(struct)
        else:
            cls_name = type(struct).__name__
            raise TypeError(f"Expected Structure, Atoms, or MSONAtoms, got {cls_name}")



# ruff: noqa: E501
from moyopy import SpaceGroupType

s = SpaceGroupType(15)  # ITA space group number (1 - 230)
assert s.hm_short == "C 2/c"
assert s.crystal_system == "Monoclinic"

print(s)
# -> PySpaceGroupType { number: 15, hm_short: "C 2/c", hm_full: "C 1 2/c 1", arithmetic_number: 8, arithmetic_symbol: "2/mC", geometric_crystal_class: "2/m", crystal_system: "Monoclinic", bravais_class: "mC", lattice_system: "Monoclinic", crystal_family: "Monoclinic" }



from __future__ import annotations

import moyopy


def test_moyo_dataset(wurtzite: moyopy.Cell):
    dataset = moyopy.MoyoDataset(wurtzite)
    assert dataset.number == 186
    assert dataset.hall_number == 480
    assert dataset.pearson_symbol == "hP4"


def test_serialization(wurtzite: moyopy.Cell):
    serialized = wurtzite.serialize_json()
    deserialized = moyopy.Cell.deserialize_json(serialized)
    assert len(wurtzite.positions) == len(deserialized.positions)


def test_moyo_dataset_repr(wurtzite: moyopy.Cell):
    dataset = moyopy.MoyoDataset(wurtzite)
    dataset_str = str(dataset)

    # Test that string representation of MoyoDataset contains key information
    assert "MoyoDataset" in dataset_str
    assert f"number={dataset.number}" in dataset_str
    assert f"hall_number={dataset.hall_number}" in dataset_str
    assert f"operations=<{len(dataset.operations)} operations>" in dataset_str
    assert f"orbits={dataset.orbits}" in dataset_str
    assert f"wyckoffs={dataset.wyckoffs}" in dataset_str

    # Test site_symmetry_symbols content without caring about quote style
    symbols = dataset.site_symmetry_symbols
    assert all(symbol in dataset_str for symbol in symbols)
    assert str(len(symbols)) in dataset_str

    # Test that repr() gives different output
    assert str(dataset) != repr(dataset)



from __future__ import annotations

import numpy as np
from pymatgen.core import Structure

import moyopy
from moyopy.interface import MoyoAdapter


class MoyoSpacegroupAnalyzer:
    def __init__(
        self,
        structure: Structure,
        symprec: float = 1e-5,
        angle_tolerance: float | None = None,
        setting: moyopy.Setting | None = None,
    ):
        self._cell = MoyoAdapter.from_structure(structure)
        self._dataset = moyopy.MoyoDataset(
            cell=self._cell,
            symprec=symprec,
            angle_tolerance=angle_tolerance,
            setting=setting,
        )

    @property
    def number(self) -> int:
        return self._dataset.number

    @property
    def hall_number(self) -> int:
        return self._dataset.hall_number

    @property
    def rotations(self) -> np.ndarray:
        return np.array(self._dataset.operations.rotations)

    @property
    def translations(self) -> np.ndarray:
        return np.array(self._dataset.operations.translations)

    @property
    def orbits(self) -> list[int]:
        return self._dataset.orbits

    @property
    def wyckoffs(self) -> list[str]:
        return self._dataset.wyckoffs

    @property
    def site_symmetry_symbols(self) -> list[str]:
        return self._dataset.site_symmetry_symbols

    @property
    def std_structure(self) -> Structure:
        return MoyoAdapter.get_structure(self._dataset.std_cell)

    @property
    def std_linear(self) -> np.ndarray:
        return np.array(self._dataset.std_linear)

    @property
    def std_origin_shift(self) -> np.ndarray:
        return np.array(self._dataset.std_origin_shift)

    @property
    def std_rotation_matrix(self) -> np.ndarray:
        return np.array(self._dataset.std_rotation_matrix)

    @property
    def prim_std_structure(self) -> Structure:
        return MoyoAdapter.get_structure(self._dataset.prim_std_cell)

    @property
    def prim_std_linear(self) -> np.ndarray:
        return np.array(self._dataset.prim_std_linear)

    @property
    def prim_std_origin_shift(self) -> np.ndarray:
        return np.array(self._dataset.prim_std_origin_shift)

    @property
    def mapping_std_prim(self) -> list[int]:
        return self._dataset.mapping_std_prim

    @property
    def symprec(self) -> float:
        return self._dataset.symprec

    @property
    def angle_tolerance(self) -> float | None:
        return self._dataset.angle_tolerance


if __name__ == "__main__":
    a = 4.0
    structure = Structure(
        lattice=np.eye(3) * a,
        species=["Al"] * 4,
        coords=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ]
        ),
    )
    msa = MoyoSpacegroupAnalyzer(structure)
    assert msa.number == 225
    assert msa.rotations.shape == (48 * 4, 3, 3)
    assert msa.orbits == [0, 0, 0, 0]
    assert msa.prim_std_structure.num_sites == 1



from moyopy._base import Operations

class Setting:
    """Preference for the setting of the space group."""
    @classmethod
    def spglib(cls) -> Setting:
        """The setting of the smallest Hall number."""
    @classmethod
    def standard(cls) -> Setting:
        """Unique axis b, cell choice 1 for monoclinic, hexagonal axes for rhombohedral,
        and origin choice 2 for centrosymmetric space groups."""
    @classmethod
    def hall_number(cls, hall_number: int) -> Setting:
        """Specific Hall number from 1 to 530."""

class Centering: ...

class HallSymbolEntry:
    """An entry containing space-group information for a specified hall_number."""
    def __init__(self, hall_number: int): ...
    @property
    def hall_number(self) -> int:
        """Number for Hall symbols (1 - 530)."""
    @property
    def number(self) -> int:
        """ITA number for space group types (1 - 230)."""
    @property
    def arithmetic_number(self) -> int:
        """Number for arithmetic crystal classes (1 - 73)."""
    @property
    def setting(self) -> Setting:
        """Setting."""
    @property
    def hall_symbol(self) -> str:
        """Hall symbol."""
    @property
    def hm_short(self) -> str:
        """Hermann-Mauguin symbol in short notation."""
    @property
    def hm_full(self) -> str:
        """Hermann-Mauguin symbol in full notation."""
    @property
    def centering(self) -> Centering:
        """Centering."""

class SpaceGroupType:
    """Space-group type information."""
    def __init__(self, number: int): ...
    # Space group type
    @property
    def number(self) -> int:
        """ITA number for space group types (1 - 230)."""
    @property
    def hm_short(self) -> str:
        """Hermann-Mauguin symbol in short notation."""
    @property
    def hm_full(self) -> str:
        """Hermann-Mauguin symbol in full notation."""
    # Arithmetic crystal class
    @property
    def arithmetic_number(self) -> int:
        """Number for arithmetic crystal classes (1 - 73)."""
    @property
    def arithmetic_symbol(self) -> str:
        """Symbol for arithmetic crystal class.

        See https://github.com/spglib/moyo/blob/main/moyo/src/data/arithmetic_crystal_class.rs
        for string values.
        """
    # Other classifications
    @property
    def geometric_crystal_class(self) -> str:
        """Geometric crystal class.

        See https://github.com/spglib/moyo/blob/main/moyo/src/data/classification.rs
        for string values.
        """
    @property
    def crystal_system(self) -> str:
        """Crystal system.

        See https://github.com/spglib/moyo/blob/main/moyo/src/data/classification.rs
        for string values.
        """
    @property
    def bravais_class(self) -> str:
        """Bravais class.

        See https://github.com/spglib/moyo/blob/main/moyo/src/data/classification.rs
        for string values.
        """
    @property
    def lattice_system(self) -> str:
        """Lattice system.

        See https://github.com/spglib/moyo/blob/main/moyo/src/data/classification.rs
        for string values.
        """
    @property
    def crystal_family(self) -> str:
        """Crystal family.

        See https://github.com/spglib/moyo/blob/main/moyo/src/data/classification.rs
        for string values.
        """

def operations_from_number(number: int, setting: Setting) -> Operations: ...



from math import sqrt

import moyopy

# https://next-gen.materialsproject.org/materials/mp-560588
a = 3.81
c = 6.24
basis = [
    [a, 0.0, 0.0],
    [-a / 2.0, a * sqrt(3.0) / 2.0, 0.0],
    [0.0, 0.0, c],
]
z1_2b = 0.00014
z2_2b = 0.37486
positions = [
    # 2b
    [1 / 3, 2 / 3, z1_2b],
    [2 / 3, 1 / 3, z1_2b + 0.5],
    # 2b
    [1 / 3, 2 / 3, z2_2b],
    [2 / 3, 1 / 3, z2_2b + 0.5],
]
numbers = [0, 0, 1, 1]
cell = moyopy.Cell(basis, positions, numbers)

dataset = moyopy.MoyoDataset(cell, symprec=1e-5, angle_tolerance=None, setting=None)
assert dataset.number == 186
assert dataset.hall_number == 480

hall_symbol_entry = moyopy.HallSymbolEntry(hall_number=dataset.hall_number)
assert hall_symbol_entry.hm_short == "P 6_3 m c"



from moyopy import SpaceGroupType


def test_space_group_type():
    s1 = SpaceGroupType(221)
    assert s1.number == 221
    assert s1.hm_short == "P m -3 m"
    assert s1.arithmetic_number == 71
    assert s1.arithmetic_symbol == "m-3mP"
    assert s1.geometric_crystal_class == "m-3m"
    assert s1.crystal_system == "Cubic"
    assert s1.bravais_class == "cP"
    assert s1.lattice_system == "Cubic"
    assert s1.crystal_family == "Cubic"

    s2 = SpaceGroupType(167)
    assert s2.number == 167
    assert s2.hm_short == "R -3 c"
    assert s2.arithmetic_number == 50
    assert s2.arithmetic_symbol == "-3mR"
    assert s2.geometric_crystal_class == "-3m"
    assert s2.crystal_system == "Trigonal"
    assert s2.bravais_class == "hR"
    assert s2.lattice_system == "Rhombohedral"
    assert s2.crystal_family == "Hexagonal"
